import {ILoggerParameters} from "./ILoggerParameters";
import {ILoggerGlobalParameters} from "./ILoggerGlobalParameters";
import {LogLevel} from "./LogLevel";
import {RecursivePartial} from "./RecursivePartial";
import * as Placeholders from 'placeholderz';

const EOL = (() => {
    let EOL;
    try{
        EOL = require('os')?.EOL;
    }catch(e){
        EOL = "\n";
    }

    return EOL;
})();

const dateformat = require('dateformat'),
    objAssign = require('assign-deep');

/**
 * @type {{displayLevels: LogLevel[], messageFormat: string, timeFormat: string, pipe: {[p: string]: "stdout" | "stderr"}}}
 */
let global_parameters: ILoggerGlobalParameters = {
    pipe: {
        [LogLevel.Debug]: 'stdout',
        [LogLevel.Info]: 'stdout',
        [LogLevel.Notice]: 'stdout',
        [LogLevel.Warning]: 'stdout',
        [LogLevel.Error]: 'stdout'
    },
    displayLevels: Object.keys(LogLevel) as LogLevel[],
    timeFormat: 'dd-mm-yyyy HH:MM:ss.l Z',
    messageFormat: '[ {time} ] [ {fmt_log_level} ] [ {fmt_module} ] {message}'
};

/**
 * @type {{messageTemplate: string}}
 */
const logger_default_parameters: ILoggerParameters = {
    messageTemplate: '{message}'
};

let longest_log_level = Math.max(...Object.values(LogLevel).map(v => v.length)),
    longest_module_name = 0;

export class Logger{

    protected readonly module_name: string;
    protected readonly params: ILoggerParameters;

    /**
     * @param {Partial<ILoggerGlobalParameters>} parameters
     */
    public static setGlobalParameters(parameters: RecursivePartial<ILoggerGlobalParameters>): void{
        global_parameters = objAssign({}, global_parameters, parameters);
    }

    /**
     * @returns {ILoggerGlobalParameters}
     */
    public static getGlobalParameters(): ILoggerGlobalParameters{
        return global_parameters;
    }

    public constructor(module_name: string, parameters?: RecursivePartial<ILoggerParameters>) {
        this.module_name = module_name;
        this.params = objAssign({}, logger_default_parameters, parameters);

        if(module_name.length > longest_module_name)
            longest_module_name = module_name.length;
    }

    /**
     * @param {LogLevel} log_level
     * @param {string} message
     * @param parameters
     */
    public write(log_level: LogLevel, message: string, ...parameters: any[]): void{
        if(global_parameters.displayLevels.indexOf(log_level) === -1)
            return;

        const placeholders = {
            time: dateformat(global_parameters.timeFormat),
            log_level: log_level.toString().toUpperCase(),
            fmt_log_level: this.format_log_level(log_level),
            module: this.module_name,
            fmt_module: this.format_module_name(this.module_name),
            message: Placeholders.replaceNamed(this.params.messageTemplate, {
                message: Placeholders.replace(message, parameters)
            })
        };

        const log_message = Placeholders.replaceNamed(global_parameters.messageFormat, placeholders);

        if(typeof process !== undefined && process !== null){
            if(process[global_parameters.pipe[log_level]] !== undefined
                && process[global_parameters.pipe[log_level]].writable){
                process[global_parameters.pipe[log_level]].write(log_message + EOL);
                return;
            }

            if(process['stdout'] !== undefined && process['stdout'].writable){
                process.stdout.write(log_message + EOL);
                return;
            }

            // Try to use console.log / console.warn / console.error
        }

        switch(log_level){
            default:
            case LogLevel.Debug:
            case LogLevel.Info:
            case LogLevel.Notice: console.log(log_message); break;
            case LogLevel.Warning: console.warn(log_message); break;
            case LogLevel.Error: console.error(log_message); break;
        }
    }

    /**
     * @param {LogLevel} log_level
     * @returns {string}
     * @protected
     */
    protected format_log_level(log_level: LogLevel): string{
        return log_level.toString().toUpperCase().padEnd(longest_log_level, ' ');
    }

    /**
     * @param {string} module_name
     * @returns {string}
     * @protected
     */
    protected format_module_name(module_name: string): string{
        return module_name.padStart(Math.round(longest_module_name/2+module_name.length/2), ' ')
            .padEnd(longest_module_name, ' ');
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public debug(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Debug, message, ...parameters);
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public info(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Info, message, ...parameters);
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public notice(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Notice, message, ...parameters);
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public warn(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Warning, message, ...parameters);
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public warning(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Warning, message, ...parameters);
    }

    /**
     * @param {string} message
     * @param parameters
     */
    public error(message: string, ...parameters: any[]): void{
        this.write(LogLevel.Error, message, ...parameters);
    }

}
